# import inspect
# import sys
# from glob import glob
# from os.path import basename, dirname, join
#
# from oidcservice.service import Service
#
#
# def factory(req_name, **kwargs):
#     pwd = dirname(__file__)
#     if pwd not in sys.path:
#         sys.path.insert(0, pwd)
#     for x in glob(join(pwd, '*.py')):
#         _mod = basename(x)[:-3]
#         if not _mod.startswith('__'):
#             # _mod = basename(x)[:-3]
#             if _mod not in sys.modules:
#                 print('"{}" not in sys.modules'.format(_mod))
#                 __import__(_mod, globals(), locals())
#
#             for name, obj in inspect.getmembers(sys.modules[_mod]):
#                 if inspect.isclass(obj) and issubclass(obj, Service):
#                     print('obj.__name__ = "{}"'.format(obj.__name__))
#                     try:
#                         if obj.__name__ == req_name:
#                             return obj(**kwargs)
#                     except AttributeError:
#                         pass
#
#     print('Failed! pwd={}, req_name={}'.format(pwd, req_name))


#

DEFAULT_SERVICES = {
    "discovery": {
        'class': 'oidcservice.oauth2.provider_info_discovery.ProviderInfoDiscovery'
    },
    'authorization': {
        'class': 'oidcservice.oauth2.authorization.Authorization'
    },
    'access_token': {
        'class': 'oidcservice.oauth2.access_token.AccessToken'
    },
    'refresh_access_token': {
        'class': 'oidcservice.oauth2.refresh_access_token.RefreshAccessToken'
    }
}



#

DEFAULT_SERVICES = {
    "discovery": {
        'class': 'oidcservice.oidc.provider_info_discovery'
                 '.ProviderInfoDiscovery'
    },
    'registration': {
        'class': 'oidcservice.oidc.registration.Registration'
    },
    'authorization': {
        'class': 'oidcservice.oidc.authorization.Authorization'
    },
    'access_token': {
        'class': 'oidcservice.oidc.access_token.AccessToken'
    },
    'refresh_access_token': {
        'class': 'oidcservice.oidc.refresh_access_token.RefreshAccessToken'
    },
    'userinfo': {
        'class': 'oidcservice.oidc.userinfo.UserInfo'
    }
}

WF_URL = "https://{}/.well-known/webfinger"
OIC_ISSUER = "http://openid.net/specs/connect/1.0/issuer"

IDT2REG = {
    'sigalg': 'id_token_signed_response_alg',
    'encalg': 'id_token_encrypted_response_alg',
    'encenc': 'id_token_encrypted_response_enc'
}

ENDPOINT2SERVICE = {
    'authorization': ['authorization'],
    'token': ['accesstoken', 'refresh_token'],
    'userinfo': ['userinfo'],
    'registration': ['registration'],
    'end_sesssion': ['end_session']
}



"""Utilities"""
import importlib
import logging
from urllib.parse import parse_qs
from urllib.parse import urlsplit
from urllib.parse import urlunsplit

import yaml
from oidcmsg.exception import UnSupported

LOGGER = logging.getLogger(__name__)

__author__ = 'roland'

URL_ENCODED = 'application/x-www-form-urlencoded'
JSON_ENCODED = "application/json"
JOSE_ENCODED = "application/jose"


def get_http_url(url, req, method='GET'):
    """
    Add a query part representing the request to a url that may already contain
    a query part. Only done if the HTTP method used is 'GET' or 'DELETE'.

    :param url: The URL
    :param req: The request as a :py:class:`oidcmsg.message.Message` instance
    :param method: The HTTP method
    :return: A possibly modified URL
    """
    if method in ["GET", "DELETE"]:
        if req.keys():
            _req = req.copy()
            comp = urlsplit(str(url))
            if comp.query:
                _req.update(parse_qs(comp.query))

            _query = str(_req.to_urlencoded())
            return urlunsplit((comp.scheme, comp.netloc, comp.path,
                               _query, comp.fragment))

        return url

    return url


def get_http_body(req, content_type=URL_ENCODED):
    """
    Get the message into the format that should be places in the body part
    of a HTTP request.

    :param req: The service request as a  :py:class:`oidcmsg.message.Message`
        instance
    :param content_type: The format of the body part.
    :return: The correctly formatet service request.
    """
    if URL_ENCODED in content_type:
        return req.to_urlencoded()

    if JSON_ENCODED in content_type:
        return req.to_json()

    if JOSE_ENCODED in content_type:
        return req  # already packaged

    raise UnSupported(
        "Unsupported content type: '%s'" % content_type)


def load_yaml_config(filename):
    """Load a YAML configuration file."""
    with open(filename, "rt", encoding='utf-8') as file:
        config_dict = yaml.safe_load(file)
    return config_dict


def modsplit(name):
    """Split importable"""
    if ':' in name:
        _part = name.split(':')
        if len(_part) != 2:
            raise ValueError("Syntax error: {s}")
        return _part[0], _part[1]

    _part = name.split('.')
    if len(_part) < 2:
        raise ValueError("Syntax error: {s}")

    return '.'.join(_part[:-1]), _part[-1]


def importer(name):
    """Import by name"""
    _part = modsplit(name)
    module = importlib.import_module(_part[0])
    return getattr(module, _part[1])



import logging

from oidcmsg import oidc
from oidcmsg.oidc import verified_claim_name
from oidcmsg.time_util import time_sans_frac

from oidcservice.oauth2 import access_token
from oidcservice.exception import ParameterError
from oidcservice.oidc import IDT2REG

__author__ = 'Roland Hedberg'

LOGGER = logging.getLogger(__name__)


class AccessToken(access_token.AccessToken):
    msg_type = oidc.AccessTokenRequest
    response_cls = oidc.AccessTokenResponse
    error_msg = oidc.ResponseMessage

    def __init__(self, service_context, state_db, client_authn_factory=None,
                 conf=None):
        access_token.AccessToken.__init__(self, service_context, state_db,
                                          client_authn_factory=client_authn_factory,
                                          conf=conf)

    def gather_verify_arguments(self):
        """
        Need to add some information before running verify()

        :return: dictionary with arguments to the verify call
        """
        _ctx = self.service_context
        # Default is RS256
        _allowed_sign_alg = _ctx.registration_response.get("id_token_signed_response_alg", "RS256")

        kwargs = {
            'client_id': _ctx.client_id, 'iss': _ctx.issuer,
            'keyjar': _ctx.keyjar, 'verify': True,
            'skew': _ctx.clock_skew, 'allowed_sign_alg': _allowed_sign_alg
        }

        for attr, param in IDT2REG.items():
            try:
                kwargs[attr] = _ctx.registration_response[param]
            except KeyError:
                pass

        try:
            kwargs['allow_missing_kid'] = self.service_context.allow[
                'missing_kid']
        except KeyError:
            pass

        return kwargs

    def update_service_context(self, resp, key='', **kwargs):
        try:
            _idt = resp[verified_claim_name('id_token')]
        except KeyError:
            pass
        else:
            try:
                if self.get_state_by_nonce(_idt['nonce']) != key:
                    raise ParameterError('Someone has messed with "nonce"')
            except KeyError:
                raise ValueError('Invalid nonce value')

            self.store_sub2state(_idt['sub'], key)

        if 'expires_in' in resp:
            resp['__expires_at'] = time_sans_frac() + int(
                resp['expires_in'])

        self.store_item(resp, 'token_response', key)

    def get_authn_method(self):
        try:
            return self.service_context.behaviour['token_endpoint_auth_method']
        except KeyError:
            return self.default_authn_method
